/**************************************************************/
/*                                                            */
/*   File name :  libvssp32.c                                 */
/*                                                            */
/*      Version:   2007-10-19                                 */
/*                                                            */
/*   Copyright (c) 2006-2007 T.Kondo/NICT All Right Reserved  */
/*                                                            */
/**************************************************************/
/*
2006-11-16
  add auxbyte1st as a member of K5VSSP_file variable
2006-09-20
  get_file_size was updated to support large file in case of VC++
2006-01-31
  libvssp32.h includes type definitions and functions 
  for utilities under src directory those use K5/VSSP32 format data.
2007-02-08
  update due to minor change of VSSP32 format as follows
    error flag added, # of bits for year is reduced from 7 to 6.
2007-02-21
  add k5vssp_t_ini function
2007-02-26
  debug in fwd1secK5
2007-03-02  
  more robust sync finding was adopted.
    1) allow 1bit pattern reversal in a sync field
    2) use sampling parameters of 1st frame as sync pattern
    3) use time information to recognize header 
2007-04-17
   consider in case of C++ 
2007-04-24
   libvssp32.h is divided into libvssp32.c and libvssp32.h
2007-10-19
  debug in fwd1seck5 (bug was introduced in 2007-02-26)
2007-10-20
  algorithm in CheckK5header is rivised for in case of that
  LLOW_USE_SAMPLE_PARAM and ALLOW_USE_SECONDS are set

*/

#include "libvssp32.h"


int64_t get_file_size(char *fname)
{
    /*  get file size opened */
	int result;
#ifdef WIN32                       //2006-09-20
    struct _stati64 info;    // for VC++    
    result=_stati64(fname,&info);  // for VC++
#else
    struct stat info;
    result=stat(fname,&info);
#endif
	if(result != 0) return (int64_t)(-1);
	else
    return (int64_t)info.st_size;
    // use %lld to print out whole number that has int64_t type
}


void vssp_aux_parse(vsspaux_t *vsspaux)
{
/*  parse AUX field data 
    [in]
      vsspaux->size  ----  AUX filed size in bytes
      vsspaux->bdat  ----  AUX field
    [return]
      vsspaux->fmt;  ----  AUX field format
      vsspaux->filter;  ----  FILTER parameter ( when fmt !=0)
      vsspaux->sid2; ----  2 char Station ID ( when fmt=1)
      vsspaux->station8;  -- station name (8 char)
      vsspaux->hostname8; -- Host PC name (8 char)



*/

    int n;
    /* fill out with space code */
    strcpy(vsspaux->sid," ");
    strcpy(vsspaux->sid2,"  ");
    strcpy(vsspaux->station8,"        ");
    strcpy(vsspaux->hostname8,"        ");
    
    /* format field get */
    vsspaux->fmt=(int)vsspaux->bdat[0];
    /* filter field get */
    vsspaux->filter=(int)vsspaux->bdat[1];
    if(vsspaux->fmt==1){
       /* station field get */
       vsspaux->sid[0]=vsspaux->bdat[1];
       vsspaux->sid[1]='\0';
       vsspaux->sid2[0]=vsspaux->bdat[2];
       vsspaux->sid2[1]=vsspaux->bdat[3];
       vsspaux->sid2[2]='\0';
       for(n=0; n<8; n++) vsspaux->station8[n]=vsspaux->bdat[4+n];
       vsspaux->station8[8]='\0';
       /* hostname field set */
       for(n=0; n<8; n++) vsspaux->hostname8[n]=vsspaux->bdat[12+n];
       vsspaux->hostname8[8]='\0';
    } else if(vsspaux->fmt==2){
       /* hostname field set */
       for(n=0; n<8; n++) vsspaux->hostname8[n]=vsspaux->bdat[12+n];
       vsspaux->hostname8[8]='\0';
    }
}

void k5vssp_t_ini(K5VSSP_file_t *K5)
{
/*
   initialize K5VSSP_file_t type variable

  */
    K5->aux32=K5->head32+3;
	//K5.auxbyte1st=0; // When info get from 1st header
	K5->auxbyte1st=20; // When compulsory set info

	K5->head16=(u_int16_t*)K5->head32;
	K5->head8=(unsigned char*)K5->head32;
    K5->twordread=0;  // reset word read counter  // maybe no more use
    K5->tbytesread=0;  // reset byte read counter
    K5->pos=0;        // reset current pointer position
	K5->framenum=0;   // frame number counter
}

int header_aux_read(K5VSSP_file_t *K5)
{
/*
  read Aux field in K5/VSSP32 header
  assumption : pointer is located just next to 64bit common header
    
  Remarks:  If K5->auxbyte1st is 0 when call this function first time,
             auxbyte information got from 1st header will be 
			 sbustituted to K5->auxbyte1st
  return  
          0: normal end
		 -99: EOF or FILE read
		  1: misalignment occurred in AUX field
		  2: other error in AUX field

      K5->ierr  0: normal end
		        -1: misalignment occurred in AUX field
		        -2: other error in AUX field

  */

       int i, i32, n;

       i32=0;

	   K5->ierr=0;
       /* next 32bit data read */
       K5->wordread = fread( K5->head32+2, sizeof( u_int32_t ), 1, K5->stream );
       //K5->twordread+=K5->wordread;
       K5->tbytesread+=K5->wordread*4;
       K5->pos+=K5->wordread*4;  // pointer position is in bytes
       if( ferror( K5->stream ) || K5->wordread==0 ){ // File EOF or read error
	       i=fclose( K5->stream ) ;
	       i32=-99;
           goto exit_here;
       }
       // parse header into each item
       K5->eflg=(int)((K5->head32[2] >> 15) & 0x1); // error flag  2007-02-07
       //K5->year=2000+(int)((K5->head32[2] >> 9) & 0x7F); // year
       K5->year=2000+(int)((K5->head32[2] >> 9) & 0x3F); // year  2007-02-07
       K5->totalday=(int)(K5->head32[2] & 0x1FF);
       K5->version1=(int)((K5->head32[2] >> 28) & 0xF);
       K5->version2=(int)((K5->head32[2] >> 24) & 0xF);
       K5->auxbyte =(int)((K5->head32[2] >> 16) & 0xFF);
       //printf("year = %d auxbyte = %d\n",K5->year,K5->auxbyte);
	   if(K5->auxbyte1st == 0) K5->auxbyte1st=K5->auxbyte;
	   if(K5->auxbyte != K5->auxbyte1st){
		   //printf("\n # of aux bytes (%d) is different from default value (%d)!\n",
           //       K5->auxbyte,K5->auxbyte1st );
#ifdef ALLOW_AUX_MISALIGN
		   // check AUX field misalignment
	       if(K5->framenum > 1){
               K5->auxbyte =(int)(K5->head32[2] & 0xFF);
			   if(K5->auxbyte == K5->auxbyte1st){  // AUX field misalignement may be occurre
                   // re-parse header into each item
                   K5->eflg=(int)((K5->head32[2] >> 15) & 0x1); 
                   K5->version1=(int)((K5->head32[2] >> 12) & 0xF);
                   K5->version2=(int)((K5->head32[2] >> 8) & 0xF);
				   i32=1;
				   K5->ierr=-1;
			   }
		   }
#endif
           //printf("Debg AUX error %d %d\n",K5->auxbyte,K5->auxbyte1st);
		   if(i32 !=1){
              K5->auxbyte=K5->auxbyte1st;
		      i32=2;
		      K5->ierr=-2; // means other error in AUX field
		   }
	   }


	   // further header read
       //for(n=0; n<K5->auxbyte/4; n++){
       for(n=0; n<K5->auxbyte1st/4; n++){  // 2007-10-19 to make robust
           K5->wordread = fread( K5->head32+3+n, sizeof( u_int32_t ), 1, K5->stream );
           //K5->twordread+=K5->wordread;
           K5->tbytesread+=K5->wordread*4;
           K5->pos+=K5->wordread*4;  // pointer position is in bytes
		   //printf("Debug word read %d  %d  %d\n",n, K5->wordread,K5->ierr);
           if( ferror( K5->stream ) || K5->wordread==0 ){ // File EOF or read error
	           i=fclose( K5->stream ) ;
		       i32=-99;
               goto exit_here;
	       }

		   if(i32==1){
              K5->eflg=(int)((K5->head32[2+K5->auxbyte/4] >> 31) & 0x1); 
              K5->year=2000+(int)((K5->head32[2+K5->auxbyte/4] >> 25) & 0x3F); 
              K5->totalday=(int)((K5->head32[2+K5->auxbyte/4] >> 16) & 0x1FF);
		   }
       }
exit_here:
       //printf("Debug %d %d\n",  i32,K5->auxbyte1st);
       return i32;

}

// K5 header check (updated from headerchkn2)
int CheckK5header(K5VSSP_file_t *K5)
/*
    K5 header common portion (1st 64bits) check

    if 1st 64bits show K5/VSSP32 pattern,  extended header and
    AUX field are read

    2007-03-02  allow at least 1 bit error in sync pattern
    2007-10-20  revise algorythm for ALLOW_USE_SECONDS and 
                ALLOW_USE_SAMPLE_PARAM cases

    return 0 : K5/VSSP Header detected
	       1 : misalignment occurred in AUX field
	       2 : other error in AUX field
		  1x : 1 bit reversal detected in sync pattern but this may be header
          -1 : K5.head32 is not header
         -99 : EOF detected or File read error

      K5->ierr  -1X:1 bit error detected in sync pattern but this may be header
*/
{
    int sdat[16]={1,2,4,8,0,0,0,0,0,0,0,0,0,0,0,0};
    int fdat[16]={40,100,200,500,1000,2000,4000,8000,16000, 
                  32000,64000,128000,256000,512000,1024000,2048000}; 
    int smplbit0, sfreq0, nch;
    u_int32_t syncpat=0x8B000000UL;    // 2nd sync pattern for K5/VSSP
    u_int32_t syncpat32=0x8C000000UL;  // 2nd sync pattern for K5/VSSP32
	u_int32_t w1, w2; //2007-03-02
	int n, nbit1, nbit2; // 2007-03-02
    int iok, iok2;
    int flag_OK_seconds,flag_OK_sample_param;

    iok=-1;

    flag_OK_seconds=1;
    flag_OK_sample_param=1;

    K5->format=0;  // suppose K5/VSSP old format

	iok2=0;
	// First 32bit sync pattern detection
    if(K5->head32[0]==0xFFFFFFFFUL) iok2=1;

    //printf("Debug %08X %08X\n",K5->head32[0],K5->head32[1]); //debug
	nbit1=0;
    nbit2=0;

#ifdef ALLOW_1BIT_REVERSE

    // allow in case of only one bit difference in FF field
	if(iok2==0){
       w1= K5->head32[0] ^ 0xFFFFFFFFUL;
	   for(n=0; n<32; n++){
	      if((w1 & 0x1UL)== 1) nbit1++;
		  w1>>= 1;
	   }
	   if(nbit1 <= 1) iok2=1;
	}
#endif
	// 2nd sync pattern detection
	if(iok2 == 1){
        if((K5->head32[1] & 0xFF000000UL) == syncpat) K5->format=1; // this is sync block
        if((K5->head32[1] & 0xFF000000UL) == syncpat32) K5->format=2; // this is sync block
        // allow in case of only one bit difference in 2nd sync field
#ifdef ALLOW_1BIT_REVERSE

	    if(K5->format==0){       // 2007-03-02
            w2 = K5->head32[1] ^ syncpat;
			w2 >>= 24;
		    for(n=0; n<8; n++){
			   if((w2 & 0x1UL)== 1) nbit2++;
			   w2>>= 1;
			}
			if(nbit2 <= 1){
				K5->format=1;
			} else {
			   nbit2=0;
               w2 = K5->head32[1] ^ syncpat32;
			   w2 >>= 24;
		       for(n=0; n<8; n++){
			      if((w2 & 0x1UL)== 1) nbit2++;
			      w2>>= 1;
			   }
			   if(nbit2 <= 1) K5->format=2;
			}
		}
#endif

	}


 
	if(K5->format > 0){ 
            iok=0;
            /* parse sampling information and get time data */ 
            K5->spatn=(int)((K5->head32[1] >> 24) & 0xFF);       
            smplbit0 = (K5->head32[1] >> 22) & 0x3;   
            K5->adbit = sdat[smplbit0];      // AD bits
            sfreq0 = ( K5->head32[1] >> 18) & 0xF;
            K5->sfreq=fdat[sfreq0];
            nch = (K5->head32[1] >> 17) & 1;
            if (nch == 1) {
                K5->numch = 4; 
            } else {
                K5->numch = 1;
            }
            K5->seconds=(long int)(K5->head32[1] & 0x1FFFFUL);
            sec2hms(K5->seconds,&K5->hh,&K5->mm,&K5->ss);
			if(K5->framenum==0){
				K5->sfreq1st=K5->sfreq;
				K5->adbit1st=K5->adbit;
				K5->numch1st=K5->numch;
				K5->seconds1st=K5->seconds;
				K5->secold=K5->seconds;
			}
			K5->framenum++;
	}

#ifdef ALLOW_USE_SECONDS
	if(K5->framenum > 1 && K5->format > 0){
        flag_OK_seconds=0;
		if((K5->seconds - K5->secold == 1) ||
			(K5->seconds - K5->secold == -86399)){   // OK
			K5->secold=K5->seconds;
            flag_OK_seconds=1;
		} 
    }
#endif

#ifdef ALLOW_USE_SAMPLE_PARAM
	if(K5->framenum > 1 && K5->format > 0){
        flag_OK_sample_param=0;
		if(K5->sfreq==K5->sfreq1st &&
			K5->adbit==K5->adbit1st &&
            K5->numch==K5->numch1st){  // OK
            flag_OK_sample_param=1;
		}
	}
#endif

    if(flag_OK_seconds == 0 && flag_OK_sample_param== 0){
	    iok=-1;
		K5->format=0;
		K5->framenum--;
    }

    if(K5->format==2){  // VSSP32 // read further header
            K5->headsizew=8;
            //if(header_aux_read(K5)<0) iok=-99;
			iok=header_aux_read(K5);
            //printf("Debug IOK %d\n",iok);
	}
	if(iok >= 0) {
		if(nbit1+nbit2 > 0){
			iok+=10;
			if(K5->ierr <= 0) K5->ierr-=10;
		}
	}

    return iok;
}

#ifdef USE_ASM

    #ifdef WIN32
        void decode_data(int *K5pos, int wd, int ilast, int pat, int adbit)  // for VC++ inline assembler
        {
        /*
                eax  --  K5pos
                ebx  --  wd
                ecx (cl)  --  adbit
                edx  --  ilast
                esi  --  loop counter
                edi  --  temporary use of wd
        
                ebp+14h --  pat

        */
                __asm{
                    mov edx, ilast
                    mov eax, K5pos
                    mov ebx, wd
                    mov cl, byte ptr [adbit]
                    //mov ch, byte ptr [pat]  //not necessary
                    xor esi,esi  
            loop_1:
                    mov edi, ebx
                    and edi, pat
                    mov [eax+esi*4], edi
                    shr ebx, cl
                    inc esi
                    cmp esi, edx
                    jnz loop_1
                }
        }
    #else  // for gcc
        void decode_data(int *K5pos, int wd, int ilast, int pat, int adbit)
        {
        /*
                K5pos --  8(%ebp)
                wd    --  12(%ebp)
                ilast --  16(%ebp)
                pat   --  20(%ebp)
                adbit --  24(%ebp)

                eax  --  K5pos
                ebx  --  wd
                ecx (cl)  --  adbit
                edx  --  ilast
                esi  --  loop counter
                edi  --  temporary use of wd
        

        */
            asm volatile (
                 "pushl %eax \n\t"     // save register used in this function
                 "pushl %ebx \n\t"
                 "pushl %ecx \n\t"
                 "pushl %edx \n\t"
                 "pushl %edi \n\t"
                 "pushl %esi \n\t"

                 "movl 16(%ebp), %edx \n\t"
                 "movl  8(%ebp), %eax \n\t"
                 "movl 12(%ebp), %ebx \n\t"

                 "movb 24(%ebp), %cl \n\t"
                 "xor %esi, %esi \n\t"  
            "loop_10: \n\t"
                 "movl %ebx, %edi \n\t"
                 "and  20(%ebp), %edi \n\t"
                 "movl %edi, (%eax,%esi,4) \n\t"
                 "shr %cl, %ebx  \n\t"
                 "inc %esi \n\t"
                 "cmp %edx, %esi \n\t"
                 "jnz loop_10 \n\t"
                 "popl %esi \n\t"    // recover saved register 
                 "popl %edi \n\t"
                 "popl %edx \n\t"
                 "popl %ecx \n\t"
                 "popl %ebx \n\t"
                 "popl %eax \n\t"
                );
        }
    #endif
#else  // normal C function
    void decode_data(int *K5pos, int wd, int ilast, int pat, int adbit)
    {
        int i,ww, wd1;
        wd1=wd;
        for(i=0; i<ilast; i++){
            ww=(int)(wd1 & pat);
            //printf("ww wd1 pat %d %d %d",ww,wd1,pat); //debug
            K5pos[i]=ww;
            wd1>>= adbit ; 
        }
    }
#endif

void parseADdata_3N(K5VSSP_file_t *K5)
{
/*
    decode K5/VSSP (IP-VLBI) data into integer data

      [in]
         K5->ubuf    ---  sampler (K5/VSSP (IP-VLBI) board) out data 
         K5->wordread --  # of data in K5->ubuf (32bit word unit)
         K5->adbit -- sampler A/D resolution bits (1,2,4,8)
      [return]
         K5->idata -- data parsed into each sampled data
         K5->npos  -- valid number of data in xdata array

     Relation between # of ch (numch) and each ch time siries

        sdata(m,k)=xdata(k*numch+m)

             where  numch :  # of channels  (1, 4)
                      m   :  CH#
                      k   :  k-th sampling
        
*/
     int numb8, pat, n, ilast, ipos;
     int pattbl[9]={1,1,3,3,15,15,15,15,255};
     u_int32_t syncpat=0x8B000000UL;    // 2nd sync pattern for K5/VSSP
     u_int32_t syncpat32=0x8C000000UL;  // 2nd sync pattern for K5/VSSP32
     u_int32_t wd1,wd2;
     int *K5pos;
     unsigned char bpat;

     if(K5->vflag != 1) goto exitend;
     //numb8=numb;
     numb8=K5->wordread*4;

     ipos=0;
     ilast=32/K5->adbit;
     pat=pattbl[K5->adbit];
     bpat=(unsigned char)pat;
     wd1=K5->ubuf[0];

     //printf("Debug IN PARSE adbit, pat %d %d\n",K5->adbit,pat);
     K5pos=K5->idata;
     for( n=1; n < K5->wordread; n++){
        wd2=K5->ubuf[n];
        /* check sync pattern */
        if ( (wd1 == 0xFFFFFFFFUL) && (((wd2 >> 24) & 0xFF) == (K5->spatn & 0xFFUL)) ) { // sync detected for K5/VSSP
            printf("Sync detected \n"); 
            continue;  // pass decoding
        }
//        for(i=0; i<ilast; i++){
//             ww=(int)(wd1 & pat);
//             //printf("ww wd1 pat %d %d %d",ww,wd1,pat); //debug
//             K5->idata[ipos]=ww;
//             wd1=wd1 >> K5->adbit ; 
//             ipos++;
//        }
        /* use of inline assembler */
        K5pos=&K5->idata[ipos];
        decode_data(K5pos, wd1, ilast, pat, K5->adbit);   // use of inline assembler
        //decode_data2(K5pos, wd1, ilast, pat, K5->adbit); // normal C function
                              // no big difference found between decode_data and decode data2
                              //   ex. decode_data  34.279 sec for 8MHz 1bit 4CH 20 sec data
                              //   ex. decode_data2 34.289 sec for 8MHz 1bit 4CH 20 sec data
                              //  direct use like below is also almost same processing speed
                              //     i.e. 34.289 sec
/*
        __asm{
            mov edx, ilast
            mov eax, K5pos
            mov ebx, wd1
            mov cl, bpat
            xor esi,esi  
    loop_1:
            mov edi, ebx
            and edi, pat
            mov [eax+esi*4], edi
            shr ebx, cl
            inc esi
            cmp esi, edx
            jnz loop_1
        }
*/
        ipos+=ilast;

        wd1=wd2;
        //K5pos+=ilast;

     }
/*     
     for(i=0; i<ilast; i++){
         ww=(int)(wd1 & pat);
         //printf("ww wd1 pat %d %d %d",ww,wd1,pat); //debug
         K5->idata[ipos]=ww;
         wd1=wd1 >> K5->adbit ; 
         ipos++;
     }
*/     

     //K5pos=K5->idata+ipos;
     K5pos=&K5->idata[ipos];
     decode_data(K5pos, wd1, ilast, pat, K5->adbit);  
     //decode_data2(K5pos, wd1, ilast, pat, K5->adbit);
     /* use of inline assembler */
/*
     __asm{
        mov edx, ilast
        mov eax, K5pos
        mov ebx, ww
        mov cl, bpat
        xor esi,esi  
loop_2:
        mov edi, ebx
        and edi, pat
        mov [eax+esi*4], edi
        shr ebx, cl
        inc esi
        cmp esi, edx
        jnz loop_2
     }
*/
     ipos+=ilast;

     K5->npos=ipos;

exitend:;
     //printf("numb8,ipos %d %d\n",numb8,ipos);
}




int CheckK5header_with_Read_org(K5VSSP_file_t *S)
{
    /* 
       Read header check to get sampling parameters 
    
       if first 64bit data is not header block, read file 
       untill finding header block.

       After call, stream pointer is set to jsut next to the header
       block.

       Return
               +ve   # of words read till header finding
               -1    no header found till EOF
    */
      
       int i, i32;
       i32=0;

       S->vflag_old=S->vflag;                  // keep current valid flag
       S->dtime_vflag_old=S->dtime_vflag;  //keep time when vflag decesion made
       S->vflag=0;

       S->headsizew=2; // for K5/VSSP
       //S->headsizew=8; // for K5/VSSP32

       /* first 32bit read */

       S->wordread = fread( S->head32, sizeof( u_int32_t ), 1, S->stream );
       S->twordread+=S->wordread;
       S->pos+=S->wordread*4;  // pointer position is in bytes
       if( ferror( S->stream ) || S->wordread==0 ){ // File EOF or read error
           i=fclose( S->stream ) ;
		   i32=-1;
           goto exit_loop;
	   }

       //printf("%X  %X\n",S->head32,S->head32+1);  //debug

loop1:
       /* next 32bit data read */
      S->wordread = fread( S->head32+1, sizeof( u_int32_t ), 1, S->stream );
      S->twordread+=S->wordread;
      S->pos+=S->wordread*4;  // pointer position is in bytes
       if( ferror( S->stream ) || S->wordread==0 ){ // File EOF or read error
	       i=fclose( S->stream ) ;
		   i32=-1;
           goto exit_loop;
	   }

       //if ((i=CheckK5header(S)) == 0) {  // sync detection
       if ((i=CheckK5header(S)) >= 0) {  // sync detection
           goto exit_loop;
       } else if(i==-99){
           i=fclose( S->stream ) ;
		   i32=-1;
           goto exit_loop;
       } else {
           S->head32[0]=S->head32[1];
           i32++;
       }
       goto loop1;
exit_loop:
       if(i32 == 0) S->vflag=1;
       S->dtime_vflag=S->seconds;
       return i32;
}


int CheckK5header_with_Read(K5VSSP_file_t *S)
{
    /* 
       Read header check to get sampling parameters 
    
       if first 64bit data is not header block, read file 
       untill finding header block.

       After call, stream pointer is set to jsut next to the header
       block.

       Return
               +ve   # of bytes read till header finding
			         (if header starts current position, then ve=0)
               -1    no header found till EOF
    */
      
       int i, i32, i8;
       i32=0;
	   i8=0;  // byte counter

       S->vflag_old=S->vflag;                  // keep current valid flag
       S->dtime_vflag_old=S->dtime_vflag;  //keep time when vflag decesion made
       S->vflag=0;

       S->headsizew=2; // for K5/VSSP
       //S->headsizew=8; // for K5/VSSP32
       /* first 7 bytes (56 bit) read */

       S->bytesread = fread( S->head8, sizeof( unsigned char ), 7, S->stream );
       S->tbytesread+=S->bytesread;
       S->pos+=S->bytesread;  // pointer position is in bytes
       if( ferror( S->stream ) || S->bytesread==0 ){ // File EOF or read error
           i=fclose( S->stream ) ;
		   i8=-1;
           goto exit_loop;
	   }
       //printf("Debug %X  %X\n",S->head32,S->head32+1);  //debug

loop1:
       /* 8 bit data read */
       S->bytesread = fread( S->head8+7, sizeof( unsigned char ), 1, S->stream );
       S->tbytesread+=S->bytesread;
       S->pos+=S->bytesread;  // pointer position is in bytes
//       if( ferror( S->stream ) || S->wordread==0 ){ // File EOF or read error
       if( ferror( S->stream ) || S->bytesread==0 ){ // File EOF or read error
	       i=fclose( S->stream ) ;
		   i8=-1;
           goto exit_loop;
	   }

//entry1:
       //if ((i=CheckK5header(S)) == 0) {  // sync detection
       if ((i=CheckK5header(S)) >= 0) {  // sync detection
           goto exit_loop;
       } else if(i==-99){
           i=fclose( S->stream ) ;
		   i8=-1;
           goto exit_loop;
       } else {
//           S->head32[0]=S->head32[1];
		   for(i=0; i<7; i++){
             S->head8[i]=S->head8[i+1];
		   }
	       i8++;
           //if(i8+2 > maxbytes) return -1;
       }
       goto loop1;
exit_loop:
       if(i8 == 0) S->vflag=1;
       S->dtime_vflag=S->seconds;
       return i8;
}


int CheckK5header_with_Read2(K5VSSP_file_t *S, int maxbytes)
{
    /* 
       Read header check to get sampling parameters 
    
       if first 64bit data is not header block, read file 
       untill finding header block.

       After call, stream pointer is set to jsut next to the header
       block.

       Return
               +ve   # of bytes read till header finding
			         (if header starts current position, then ve=0)
               -1    no header found till EOF
    */
      
       int i, i32, i8;
       i32=0;
	   i8=0;  // byte counter

       S->vflag_old=S->vflag;                  // keep current valid flag
       S->dtime_vflag_old=S->dtime_vflag;  //keep time when vflag decesion made
       S->vflag=0;

       S->headsizew=2; // for K5/VSSP
       //S->headsizew=8; // for K5/VSSP32

       /* first 7 bytes (56 bit) read */

       S->bytesread = fread( S->head8, sizeof( unsigned char ), 7, S->stream );
       S->tbytesread+=S->bytesread;
       S->pos+=S->bytesread;  // pointer position is in bytes
       if( ferror( S->stream ) || S->bytesread==0 ){ // File EOF or read error
           i=fclose( S->stream ) ;
		   i8=-1;
           goto exit_loop;
	   }
       //printf("%X  %X\n",S->head32,S->head32+1);  //debug

loop1:
       /* 8 bit data read */
       S->bytesread = fread( S->head8+7, sizeof( unsigned char ), 1, S->stream );
       S->tbytesread+=S->bytesread;
       S->pos+=S->bytesread;  // pointer position is in bytes
//       if( ferror( S->stream ) || S->wordread==0 ){ // File EOF or read error
       if( ferror( S->stream ) || S->bytesread==0 ){ // File EOF or read error
	       i=fclose( S->stream ) ;
		   i8=-1;
           goto exit_loop;
	   }

//entry1:
       //if ((i=CheckK5header(S)) == 0) {  // sync detection
       if ((i=CheckK5header(S)) >= 0) {  // sync detection
           goto exit_loop;
       } else if(i==-99){
           i=fclose( S->stream ) ;
		   i8=-1;
           goto exit_loop;
       } else {
//           S->head32[0]=S->head32[1];
		   for(i=0; i<7; i++){
             S->head8[i]=S->head8[i+1];
		   }
	       i8++;
           if(i8+2 > maxbytes) return -1;
       }
       goto loop1;
exit_loop:
       if(i8 == 0) S->vflag=1;
       S->dtime_vflag=S->seconds;
       return i8;
}


int fwdNsecK5(int nsec, K5VSSP_file_t *K5)
{
    /* 
       forward fast N sec then check header
       if header doesn't exist, rewind RWIND_BYTE data at first,
       then read data slowly to find out header.
       If still header doesn't exist, rewind 1sec data
       then read data slowly to find out header.

       Note: use of fsetpos() and fgetpos() was obsoleted because of
       the compatiblity ploblem of a fpos_t type variable between linux 2.6
       and previous one

    */
    int i,k;
    //int maxword, maxbytes;
    int maxbytes;
    int nbytes, nback;
    //int64_t tword, tbytes;
    int64_t tbytes;
	int64_t pos;

    if(nsec < 1) return 0; // nothing to do

    //tword=K5->twordread;  // save current word read counter data 
    tbytes=K5->tbytesread;  // save current byte read counter data 
    pos=K5->pos;  // save current position

    k=0;

    nbytes=K5->bytesec*nsec;
    K5->pos+=nbytes;  // pointer position to be set
    if(K5->pos < K5->size){  // not yet reached to EOF
         fseek(K5->stream,nbytes,SEEK_CUR); // move pointer by K5->bytesec
         if( ferror( K5->stream ) ){
            return -1;
         }
    } else {  // exceed EOF so pinter set to EOF
         fseek(K5->stream,(int)(K5->size - pos),SEEK_CUR); // 
         if( ferror( K5->stream ) ){
            return -1;
         }
        //K5->wordread=(int)((K5->size - pos)/4); // convert to 32bit-word unit
        K5->bytesread=(int)(K5->size - pos); 
        //K5->twordread+=(int64_t)K5->wordread;
        K5->tbytesread+=(int64_t)K5->bytesread;
        //K5->pos+=K5->wordread*4;  // pointer position is in bytes
        K5->pos+=K5->bytesread;  // pointer position is in bytes
        return -1;
    }
    //K5->wordread=nbytes/4;
    //K5->twordread+=(int64_t)K5->wordread;
    K5->bytesread=nbytes;
    //K5->bytesread+=(int64_t)K5->bytesread;  //2007-02-12 bug fix
    K5->tbytesread+=(int64_t)K5->bytesread;  // 2007-02-12 bug fix


    /* read 64 bit header */
    K5->wordread = fread(K5->head32, sizeof(u_int32_t), 2, K5->stream);
    //K5->twordread+=K5->wordread;
    K5->tbytesread+=K5->wordread*4;
    K5->pos+=K5->wordread*4;  // pointer position is in bytes
    if( ferror( K5->stream ) || K5->wordread==0 ){ // EOF or File read error
	   return -1;
    }
    if((i=CheckK5header(&(*K5))) < 0){  // this is not the case of header
       if(i==-99) return -1;
       /* rewind file */
       if(REWIND_BYTE == 0){
           fseek(K5->stream,-(int)(K5->pos-pos),SEEK_CUR); // set position at initial position
           //K5->twordread=tword;            // word counter retrieve
           K5->tbytesread=tbytes;            // byte counter retrieve
           K5->pos=pos;  // retrieve initial position
           if((k=CheckK5header_with_Read(&(*K5))) < 0){ // EOF or file read error
               return -1;
           }
       } else {
           //maxword=REWIND_BYTE/4;
           maxbytes=REWIND_BYTE;
           fseek(K5->stream,-(int)REWIND_BYTE,SEEK_CUR); // set position at initial position
           //K5->twordread-=REWIND_BYTE/4;            // word counter retrieve
           K5->tbytesread-=REWIND_BYTE;            // word counter retrieve
           K5->pos-=REWIND_BYTE;  // retrieve initial position
           /* 1st trial */
           //if((k=CheckK5header_with_Read2(&(*K5),maxword)) < 0){ // EOF or file read error
           if((k=CheckK5header_with_Read2(&(*K5),maxbytes)) < 0){ // EOF or file read error
               /* 2nd trial */
               if(nsec==1){
                   nback=(int)(K5->pos-pos);  // rewind to initial position
               } else {
                   nback=K5->bytesec;   // rewind just 1 sec
               }

               fseek(K5->stream,-nback,SEEK_CUR); // set position at initial position
               //K5->twordread-=nback/4;            // word counter rewind
               K5->tbytesread-=nback;            // word counter rewind
               K5->pos-=nback;  // initial position rewind
               if((k=CheckK5header_with_Read(&(*K5))) < 0){ // EOF or file read error
                   return -1;
               }
           }
       }
       // get headder after k bytes read
    }
       
    return k;
}

int fwd1secK5(K5VSSP_file_t *K5)
{
    /* 

      2007-10-19

       forward fast 1 sec then check header
       if header doesn't exist, rewind RWIND_BYTE data at first,
       then read data slowly to find out header.
       If still header doesn't exist, rewind previous header position
       then read data slowly to find out header.

       Note: use of fsetpos() and fgetpos() was obsoleted because of
       the compatiblity ploblem of a fpos_t type variable between linux 2.6
       and previous one

   Return
      fwd1secK5   0: not slip occurred
				 -1:  error or EOF

        K5->pos  :  current pointer position in a file
		K5->tbytesread : total bytes read till header found

				   


    */
    int i,k;
    //int maxword, maxbytes;
    int maxbytes;
    //int64_t tword, tbytes;
    int64_t tbytes;
	int64_t pos;
    //tword=K5->twordread;  // save current word read counter data 
    tbytes=K5->tbytesread;  // save current byte read counter data 
    pos=K5->pos;  // save current position

    k=0;

    K5->pos+=K5->bytesec;  // pointer position to be set
    if(K5->pos < K5->size){  // not yet reached to EOF
         fseek(K5->stream,K5->bytesec,SEEK_CUR); // move pointer by K5->bytesec
         if( ferror( K5->stream ) ){
            return -1;
         }
    } else {  // exceed EOF so pinter set to EOF
         fseek(K5->stream,(int)(K5->size - pos),SEEK_CUR); // 
         if( ferror( K5->stream ) ){
            return -1;
         }
        //K5->wordread=(int)((K5->size - pos)/4); // convert to 32bit-word unit
        K5->bytesread=(int)(K5->size - pos); 
        //K5->twordread+=(int64_t)K5->wordread;
        K5->tbytesread+=(int64_t)K5->bytesread;
        //K5->pos+=K5->wordread*4;  // pointer position is in bytes
        K5->pos+=K5->bytesread;  // pointer position is in bytes
        return -1;
    }
    //K5->wordread=K5->bytesec/4;
    K5->bytesread=K5->bytesec;
    //K5->twordread+=(int64_t)K5->wordread;
    K5->tbytesread+=(int64_t)K5->bytesread;

    /* read 64 bit header */
    K5->wordread = fread(K5->head32, sizeof(u_int32_t), 2, K5->stream);
    //K5->twordread+=K5->wordread;
    K5->tbytesread+=K5->wordread*4;
    K5->pos+=K5->wordread*4;  // pointer position is in bytes
    if( ferror( K5->stream ) || K5->wordread==0 ){ // EOF or File read error
	   return -1;
    }
    if((i=CheckK5header(&(*K5))) < 0){  // this is not the case of header
       if(i==-99) return -1;
       /* rewind file */
       if(REWIND_BYTE == 0){
           fseek(K5->stream,-(int)(K5->pos-pos),SEEK_CUR); // set position at initial position
           K5->tbytesread=tbytes;            // byte counter retrieve
           K5->pos=pos;  // retrieve initial position
           if((k=CheckK5header_with_Read(&(*K5))) < 0){ // EOF or file read error
               return -1;
           }
       } else {
           maxbytes=REWIND_BYTE;
           // set file pointer to current poniter - REWIND_BYTE
           fseek(K5->stream,-(int)REWIND_BYTE,SEEK_CUR); // set position at initial position

           //K5->twordread-=REWIND_BYTE/4;            // word counter retrieve
           K5->tbytesread-=REWIND_BYTE;            // byte counter retrieve
           K5->pos-=REWIND_BYTE;  //   this is current file pointer (set A)
           /* 1st trial */
           //if((k=CheckK5header_with_Read2(&(*K5),maxword)) < 0){ // EOF or file read error
           if((k=CheckK5header_with_Read2(&(*K5),2*maxbytes)) < 0){ // EOF or file read error
               // now file pointer is A+2*maxbytes
               /* 2nd trial */
               fseek(K5->stream,-(int)(K5->pos-pos),SEEK_CUR); // set position at initial position
               K5->tbytesread=tbytes;            // byte counter retrieve
         //      K5->tbytesread=tbytes+maxbytes;            // byte counter retrieve
               K5->pos=pos;  // retrieve initial position
         //      K5->pos=pos+maxbytes;  // retrieve initial position
               if((k=CheckK5header_with_Read(&(*K5))) < 0){ // EOF or file read error
                   return -1;
               }
			   //printf("debug\n");
           }
       }
       // get headder after k bytes read
    }
       
    return k;
}

